"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os
import re
import logging
import numpy as np
import matplotlib.pyplot as plt
from tensorflow.python.summary.summary_iterator import summary_iterator

# project imports
from lamareg.SynthSeg.predict import predict

# third-party imports
from lamareg.ext.lab2im import utils


def validate_training(
    image_dir,
    gt_dir,
    models_dir,
    validation_main_dir,
    labels_segmentation,
    n_neutral_labels=None,
    evaluation_labels=None,
    step_eval=1,
    min_pad=None,
    cropping=None,
    target_res=1.0,
    gradients=False,
    flip=False,
    topology_classes=None,
    sigma_smoothing=0,
    keep_biggest_component=False,
    n_levels=5,
    nb_conv_per_level=2,
    conv_size=3,
    unet_feat_count=24,
    feat_multiplier=2,
    activation="elu",
    recompute=False,
):
    """This function validates models saved at different epochs of the same training.
    All models are assumed to be in the same folder.
    The results of each model are saved in a subfolder in validation_main_dir.
    :param image_dir: path of the folder with validation images.
    :param gt_dir: path of the folder with ground truth label maps.
    These are matched to the validation images by sorting order.
    :param models_dir: path of the folder with the models to validate.
    :param validation_main_dir: path of the folder where all the models validation subfolders will be saved.
    :param labels_segmentation: path of the numpy array containing all the segmentation labels used during training.
    :param n_neutral_labels: (optional) value of n_neutral_labels used during training. Used only if flip is True.
    :param evaluation_labels: (optional) label values to validate on. Must be a subset of the segmentation labels.
    Can be a sequence, a 1d numpy array, or the path to a numpy 1d array. Default is the same as segmentation_label_list
    :param step_eval: (optional) If step_eval > 1 skips models when validating, by validating on models step_eval apart.
    :param cropping: (optional) whether to crop the input to smaller size while being run through the network.
    The result is then given in the original image space. Can be an int, a sequence, or a 1d numpy array.
    :param target_res: (optional) target resolution at which the network operates (and thus resolution of the output
    segmentations). This must match the resolution of the training data ! target_res is used to automatically resampled
    the images with resolutions outside [target_res-0.05, target_res+0.05].
    Can be a sequence, a 1d numpy array. Set to None to disable the automatic resampling. Default is 1mm.
    :param flip: (optional) whether to perform test-time augmentation, where the input image is segmented along with
    a right/left flipped version on it. If set to True (default), be careful because this requires more memory.
    :param topology_classes: List of classes corresponding to all segmentation labels, in order to group them into
    classes, for each of which we will operate a smooth version of biggest connected component.
    Can be a sequence, a 1d numpy array, or the path to a numpy 1d array in the same order as segmentation_label_list.
    Default is None, where no topological analysis is performed.
    :param sigma_smoothing: (optional) If not None, the posteriors are smoothed with a gaussian kernel of the specified
    standard deviation.
    :param keep_biggest_component: (optional) whether to only keep the biggest component in the predicted segmentation.
    This is applied independently of topology_classes, and it is applied to the whole segmentation
    :param n_levels: (optional) number of level for the Unet. Default is 5.
    :param nb_conv_per_level: (optional) number of convolutional layers per level. Default is 2.
    :param conv_size: (optional) size of the convolution kernels. Default is 2.
    :param unet_feat_count: (optional) number of feature maps for the first level. Default is 24.
    :param feat_multiplier: (optional) multiply the number of feature by this number at each new level. Default is 1.
    :param activation: (optional) activation function. Can be 'elu', 'relu'.
    :param recompute: (optional) whether to recompute result files even if they already exist.
    """

    # create result folder
    utils.mkdir(validation_main_dir)

    # loop over models
    list_models = utils.list_files(models_dir, expr=["dice", ".h5"], cond_type="and")[
        ::step_eval
    ]
    # list_models = [p for p in list_models if int(os.path.basename(p)[-6:-3]) % 10 == 0]
    loop_info = utils.LoopInfo(len(list_models), 1, "validating", True)
    for model_idx, path_model in enumerate(list_models):

        # build names and create folders
        model_val_dir = os.path.join(
            validation_main_dir, os.path.basename(path_model).replace(".h5", "")
        )
        dice_path = os.path.join(model_val_dir, "dice.npy")
        utils.mkdir(model_val_dir)

        if (not os.path.isfile(dice_path)) | recompute:
            loop_info.update(model_idx)
            predict(
                path_images=image_dir,
                path_segmentations=model_val_dir,
                path_model=path_model,
                labels_segmentation=labels_segmentation,
                n_neutral_labels=n_neutral_labels,
                min_pad=min_pad,
                cropping=cropping,
                target_res=target_res,
                gradients=gradients,
                flip=flip,
                topology_classes=topology_classes,
                sigma_smoothing=sigma_smoothing,
                keep_biggest_component=keep_biggest_component,
                n_levels=n_levels,
                nb_conv_per_level=nb_conv_per_level,
                conv_size=conv_size,
                unet_feat_count=unet_feat_count,
                feat_multiplier=feat_multiplier,
                activation=activation,
                gt_folder=gt_dir,
                evaluation_labels=evaluation_labels,
                recompute=recompute,
                verbose=False,
            )


def plot_validation_curves(
    list_validation_dirs,
    architecture_names=None,
    eval_indices=None,
    skip_first_dice_row=True,
    size_max_circle=100,
    figsize=(11, 6),
    y_lim=None,
    fontsize=18,
    list_linestyles=None,
    list_colours=None,
    plot_legend=False,
    draw_line=None,
):
    """This function plots the validation curves of several networks, based on the results of validate_training().
    It takes as input a list of validation folders (one for each network), each containing subfolders with dice scores
    for the corresponding validated epoch.
    :param list_validation_dirs: list of all the validation folders of the trainings to plot.
    :param eval_indices: (optional) compute the average Dice on a subset of labels indicated by the specified indices.
    Can be a 1d numpy array, the path to such an array, or a list of 1d numpy arrays as long as list_validation_dirs.
    :param skip_first_dice_row: if eval_indices is None, skip the first row of the dice matrices (usually background)
    :param size_max_circle: (optional) size of the marker for epochs achieving the best validation scores.
    :param figsize: (optional) size of the figure to draw.
    :param fontsize: (optional) fontsize used for the graph."""

    n_curves = len(list_validation_dirs)

    if eval_indices is not None:
        if isinstance(eval_indices, (np.ndarray, str)):
            if isinstance(eval_indices, str):
                eval_indices = np.load(eval_indices)
            eval_indices = np.squeeze(
                utils.reformat_to_n_channels_array(
                    eval_indices, n_dims=len(eval_indices)
                )
            )
            eval_indices = [eval_indices] * len(list_validation_dirs)
        elif isinstance(eval_indices, list):
            for i, e in enumerate(eval_indices):
                if isinstance(e, np.ndarray):
                    eval_indices[i] = np.squeeze(
                        utils.reformat_to_n_channels_array(e, n_dims=len(e))
                    )
                else:
                    raise TypeError(
                        "if provided as a list, eval_indices should only contain numpy arrays"
                    )
        else:
            raise TypeError(
                "eval_indices can be a numpy array, a path to a numpy array, or a list of numpy arrays."
            )
    else:
        eval_indices = [None] * len(list_validation_dirs)

    # reformat model names
    if architecture_names is None:
        architecture_names = [
            os.path.basename(os.path.dirname(d)) for d in list_validation_dirs
        ]
    else:
        architecture_names = utils.reformat_to_list(
            architecture_names, len(list_validation_dirs)
        )

    # prepare legend labels
    if plot_legend is False:
        list_legend_labels = ["_nolegend_"] * n_curves
    elif plot_legend is True:
        list_legend_labels = architecture_names
    else:
        list_legend_labels = architecture_names
        list_legend_labels = [
            "_nolegend_" if i >= plot_legend else list_legend_labels[i]
            for i in range(n_curves)
        ]

    # prepare linestyles
    if list_linestyles is not None:
        list_linestyles = utils.reformat_to_list(list_linestyles)
    else:
        list_linestyles = [None] * n_curves

    # prepare curve colours
    if list_colours is not None:
        list_colours = utils.reformat_to_list(list_colours)
    else:
        list_colours = [None] * n_curves

    # loop over architectures
    plt.figure(figsize=figsize)
    for idx, (
        net_val_dir,
        net_name,
        linestyle,
        colour,
        legend_label,
        eval_idx,
    ) in enumerate(
        zip(
            list_validation_dirs,
            architecture_names,
            list_linestyles,
            list_colours,
            list_legend_labels,
            eval_indices,
        )
    ):

        list_epochs_dir = utils.list_subfolders(net_val_dir, whole_path=False)

        # loop over epochs
        list_net_scores = list()
        list_epochs = list()
        for epoch_dir in list_epochs_dir:

            # build names and create folders
            path_epoch_scores = os.path.join(net_val_dir, epoch_dir, "dice.npy")
            if os.path.isfile(path_epoch_scores):
                if eval_idx is not None:
                    list_net_scores.append(
                        np.mean(np.abs(np.load(path_epoch_scores)[eval_idx, :]))
                    )
                else:
                    if skip_first_dice_row:
                        list_net_scores.append(
                            np.mean(np.abs(np.load(path_epoch_scores)[1:, :]))
                        )
                    else:
                        list_net_scores.append(
                            np.mean(np.abs(np.load(path_epoch_scores)))
                        )
                list_epochs.append(int(re.sub("[^0-9]", "", epoch_dir)))

        # plot validation scores for current architecture
        if list_net_scores:  # check that archi has been validated for at least 1 epoch
            list_net_scores = np.array(list_net_scores)
            list_epochs = np.array(list_epochs)
            list_epochs, idx = np.unique(list_epochs, return_index=True)
            list_net_scores = list_net_scores[idx]
            max_score = np.max(list_net_scores)
            epoch_max_score = list_epochs[np.argmax(list_net_scores)]
            print("\n" + net_name)
            print("epoch max score: %d" % epoch_max_score)
            print("max score: %0.3f" % max_score)
            plt.plot(
                list_epochs,
                list_net_scores,
                label=legend_label,
                linestyle=linestyle,
                color=colour,
            )
            plt.scatter(epoch_max_score, max_score, s=size_max_circle, color=colour)

    # finalise plot
    plt.grid()
    if draw_line is not None:
        draw_line = utils.reformat_to_list(draw_line)
        list_linestyles = ["dotted", "dashed", "solid", "dashdot"][: len(draw_line)]
        for line, linestyle in zip(draw_line, list_linestyles):
            plt.axhline(line, color="black", linestyle=linestyle)
    plt.tick_params(axis="both", labelsize=fontsize)
    plt.ylabel("Scores", fontsize=fontsize)
    plt.xlabel("Epochs", fontsize=fontsize)
    if y_lim is not None:
        plt.ylim(y_lim[0], y_lim[1] + 0.01)  # set right/left limits of plot
    plt.title("Validation curves", fontsize=fontsize)
    if plot_legend:
        plt.legend(fontsize=fontsize)
    plt.tight_layout(pad=1)
    plt.show()


def draw_learning_curve(
    path_tensorboard_files,
    architecture_names,
    figsize=(11, 6),
    fontsize=18,
    y_lim=None,
    remove_legend=False,
):
    """This function draws the learning curve of several trainings on the same graph.
    :param path_tensorboard_files: list of tensorboard files corresponding to the models to plot.
    :param architecture_names: list of the names of the models
    :param figsize: (optional) size of the figure to draw.
    :param fontsize: (optional) fontsize used for the graph.
    """

    # reformat inputs
    path_tensorboard_files = utils.reformat_to_list(path_tensorboard_files)
    architecture_names = utils.reformat_to_list(architecture_names)
    assert len(path_tensorboard_files) == len(
        architecture_names
    ), "names and tensorboard lists should have same length"

    # loop over architectures
    plt.figure(figsize=figsize)
    for path_tensorboard_file, name in zip(path_tensorboard_files, architecture_names):

        path_tensorboard_file = utils.reformat_to_list(path_tensorboard_file)

        # extract loss at the end of all epochs
        list_losses = list()
        list_epochs = list()
        logging.getLogger("tensorflow").disabled = True
        for path in path_tensorboard_file:
            for e in summary_iterator(path):
                for v in e.summary.value:
                    if v.tag == "loss" or v.tag == "accuracy" or v.tag == "epoch_loss":
                        list_losses.append(v.simple_value)
                        list_epochs.append(e.step)
        plt.plot(
            np.array(list_epochs), 1 - np.array(list_losses), label=name, linewidth=2
        )

    # finalise plot
    plt.grid()
    if not remove_legend:
        plt.legend(fontsize=fontsize)
    plt.xlabel("Epochs", fontsize=fontsize)
    plt.ylabel("Scores", fontsize=fontsize)
    if y_lim is not None:
        plt.ylim(y_lim[0], y_lim[1] + 0.01)  # set right/left limits of plot
    plt.tick_params(axis="both", labelsize=fontsize)
    plt.title("Learning curves", fontsize=fontsize)
    plt.tight_layout(pad=1)
    plt.show()



"""
coregister - Image Registration for Aligning Neuroimaging Data

Part of the micaflow processing pipeline for neuroimaging data.

This module performs comprehensive image registration between two images using the
Advanced Normalization Tools (ANTs) SyNRA algorithm, which combines rigid, affine,
and symmetric normalization transformations. It aligns a moving image with a fixed
reference space, enabling spatial normalization of neuroimaging data for group analysis,
multimodal integration, or atlas-based analyses.

Features:
--------
- Combined rigid, affine, and SyN nonlinear registration in one step
- Bidirectional transformation capability (forward and inverse)
- Option to save all transformation components for later application
- Uses ANTs' powerful SyNRA algorithm for optimal accuracy
- Preserves header information in the registered output images

API Usage:
---------
micaflow coregister
    --fixed-file <path/to/reference.nii.gz>
    --moving-file <path/to/source.nii.gz>
    --output <path/to/registered.nii.gz>
    [--warp-file <path/to/warp.nii.gz>]
    [--affine-file <path/to/affine.mat>]
    [--rev-warp-file <path/to/reverse_warp.nii.gz>]
    [--rev-affine-file <path/to/reverse_affine.mat>]

Python Usage:
-----------
>>> from micaflow.scripts.coregister import ants_linear_nonlinear_registration
>>> ants_linear_nonlinear_registration(
...     fixed_file="mni152.nii.gz",
...     moving_file="subject_t1w.nii.gz",
...     out_file="registered_t1w.nii.gz",
...     warp_file="warp.nii.gz",
...     affine_file="affine.mat",
...     rev_warp_file="reverse_warp.nii.gz",
...     rev_affine_file="reverse_affine.mat"
... )

"""

import ants
import argparse
import shutil
import sys
from colorama import init, Fore, Style
import os
import multiprocessing
import nibabel as nib
import numpy as np
from scipy.interpolate import RegularGridInterpolator
from concurrent.futures import ThreadPoolExecutor
import tempfile

init()

# Get number of available CPU cores
DEFAULT_THREADS = multiprocessing.cpu_count()


def print_help():
    """Print a help message with examples."""
    # ANSI color codes
    CYAN = Fore.CYAN
    GREEN = Fore.GREEN
    YELLOW = Fore.YELLOW
    BLUE = Fore.BLUE
    MAGENTA = Fore.MAGENTA
    BOLD = Style.BRIGHT
    RESET = Style.RESET_ALL

    help_text = f"""
    {CYAN}{BOLD}╔════════════════════════════════════════════════════════════════╗
    ║                      IMAGE COREGISTRATION                      ║
    ╚════════════════════════════════════════════════════════════════╝{RESET}

    This script performs linear (rigid + affine) and nonlinear (SyN) registration 
    between two images using ANTs. The registration aligns the moving image to 
    match the fixed reference image space.

    {CYAN}{BOLD}────────────────────────── USAGE ──────────────────────────{RESET}
      micaflow coregister {GREEN}[options]{RESET}

    {CYAN}{BOLD}─────────────────── REQUIRED ARGUMENTS ───────────────────{RESET}
      {YELLOW}--fixed{RESET}   : Path to the fixed/reference image (.nii.gz)
      {YELLOW}--moving{RESET}  : Path to the moving image to be registered (.nii.gz)

    {CYAN}{BOLD}─────────────────── OPTIONAL ARGUMENTS ───────────────────{RESET}
      {YELLOW}--output{RESET}             : Output path for the registered image (.nii.gz)
      {YELLOW}--registration-method{RESET}: Registration method (default: "SyNRA")
      {YELLOW}--warp-file{RESET}          : Path to save the forward warp field (.nii.gz)
      {YELLOW}--affine-file{RESET}        : Path to save the forward affine transform (.mat)
      {YELLOW}--inverse-warp-file{RESET}  : Path to save the reverse warp field (.nii.gz)
      {YELLOW}--initial-affine-file{RESET}: Path to initial affine transform to use (.mat)
      {YELLOW}--initial-warp-file{RESET}  : Path to initial warp field to use (.nii.gz)
      {YELLOW}--interpolator{RESET}       : Interpolation method (default: "genericLabel")
      {YELLOW}--threads{RESET}            : Number of CPU threads (default: all available)

    {CYAN}{BOLD}───────────────── ANTS REGISTRATION OPTIONS ────────────────{RESET}
      {YELLOW}--verbose{RESET}             : Enable verbose output
      {YELLOW}--grad-step{RESET}           : Gradient step size (default: 0.2)
      {YELLOW}--flow-sigma{RESET}          : Smoothing for update field (default: 3)
      {YELLOW}--total-sigma{RESET}         : Smoothing for total field (default: 0)
      {YELLOW}--aff-metric{RESET}          : Metric for affine stage (default: "mattes")
      {YELLOW}--aff-sampling{RESET}        : Sampling parameter for affine metric (default: 32)
      {YELLOW}--syn-metric{RESET}          : Metric for SyN stage (default: "mattes")
      {YELLOW}--syn-sampling{RESET}        : Sampling parameter for SyN metric (default: 32)
      {YELLOW}--reg-iterations{RESET}      : SyN iterations, comma-separated (e.g., "40,20,0")
      {YELLOW}--aff-iterations{RESET}      : Affine iterations, comma-separated (e.g., "2100,1200,1200,10")
      {YELLOW}--aff-shrink-factors{RESET}  : Affine shrink factors, comma-separated (e.g., "6,4,2,1")
      {YELLOW}--aff-smoothing-sigmas{RESET}: Affine smoothing sigmas, comma-separated (e.g., "3,2,1,0")
      {YELLOW}--random-seed{RESET}         : Random seed for reproducibility

    {CYAN}{BOLD}────────────────── EXAMPLE USAGE ────────────────────────{RESET}

    {BLUE}# Basic registration with default parameters{RESET}
    micaflow coregister {YELLOW}--fixed{RESET} mni152.nii.gz {YELLOW}--moving{RESET} subject_t1w.nii.gz \\
      {YELLOW}--output{RESET} registered_t1w.nii.gz

    {BLUE}# Registration with saved transforms{RESET}
    micaflow coregister {YELLOW}--fixed{RESET} mni152.nii.gz {YELLOW}--moving{RESET} subject_t1w.nii.gz \\
      {YELLOW}--output{RESET} registered_t1w.nii.gz {YELLOW}--warp-file{RESET} warp.nii.gz {YELLOW}--affine-file{RESET} affine.mat

    {BLUE}# Registration with custom ANTs parameters{RESET}
    micaflow coregister {YELLOW}--fixed{RESET} mni152.nii.gz {YELLOW}--moving{RESET} subject_t1w.nii.gz \\
      {YELLOW}--output{RESET} registered_t1w.nii.gz {YELLOW}--reg-iterations{RESET} "100,50,20" \\
      {YELLOW}--grad-step{RESET} 0.1 {YELLOW}--verbose{RESET}

    {CYAN}{BOLD}────────────────────────── NOTES ───────────────────────{RESET}
    {MAGENTA}•{RESET} The registration performs SyNRA transformation (rigid+affine+SyN)
    {MAGENTA}•{RESET} Forward transforms convert from moving space to fixed space
    {MAGENTA}•{RESET} Reverse transforms convert from fixed space to moving space
    {MAGENTA}•{RESET} The transforms can be applied to other images using apply_warp
    {MAGENTA}•{RESET} For reproducible results, set the random seed
    """
    print(help_text)

def resample_displacement_field_vectorially(warp_img, target_img, interp='bSpline'):
    """
    Resample a displacement field from warp_img space to target_img space
    treating it as a vector field in physical mm coordinates.

    Parameters
    ----------
    warp_img : ants.ANTsImage
        Original displacement field (LPS mm, 3 components).
    target_img : ants.ANTsImage
        Target image defining the new grid.
    interp : str
        Interpolation type ('linear' or 'bSpline' are good choices).

    Returns
    -------
    ants.ANTsImage
        Displacement field resampled to target grid, in mm (LPS).
    """
    # Get ANTs direction matrices and spacing
    src_direction = np.array(warp_img.direction).reshape(3, 3)
    tgt_direction = np.array(target_img.direction).reshape(3, 3)
    src_spacing = np.array(warp_img.spacing)
    tgt_spacing = np.array(target_img.spacing)
    
    # Construct affine-like matrices (direction * spacing)
    src_affine = src_direction * src_spacing
    tgt_affine = tgt_direction * tgt_spacing

    # Get vector components
    components = ants.split_channels(warp_img)
    n_comp = len(components)
    if n_comp != 3:
        raise ValueError(f"Expected 3 components, got {n_comp}")

    # Prepare target grid - resample each component as scalar
    target_components = []
    for i in range(n_comp):
        resampled = ants.resample_image_to_target(
            components[i], target_img, interp_type=interp
        )
        target_components.append(resampled)

    # Merge back into 3-channel vector field
    resampled_field = ants.merge_channels(target_components)

    # Rotate vectors from source frame into target frame
    R = np.dot(tgt_affine, np.linalg.inv(src_affine))
    data = resampled_field.numpy()  # shape (X,Y,Z,3)

    # Apply rotation to each vector voxel
    X, Y, Z, _ = data.shape
    data_flat = data.reshape(-1, 3)
    data_rot = (data_flat @ R.T).reshape(X, Y, Z, 3)

    # Create new ANTs image with rotated vectors
    # We need to create it component by component and then merge
    rotated_components = []
    for i in range(3):
        comp = ants.from_numpy(
            data_rot[..., i],
            origin=target_img.origin,
            spacing=target_img.spacing,
            direction=target_img.direction
        )
        rotated_components.append(comp)
    
    # Merge into multi-component image
    out = ants.merge_channels(rotated_components)

    return out

def compose_3d_fields_extrap(A, B, interpolation_order=3):
    """
    Compose two 3D displacement fields A and B using high-order interpolation
    with linear extrapolation at boundaries.

    Mathematical Framework:
    ----------------------
    Given displacement fields A and B (both in voxel units), this function computes
    their composition C, which represents applying transformation A first, then B.
    
    For a displacement field D, the induced deformation mapping is:
        T_D(x) = x + D(x)
    
    The composition T = T_B ∘ T_A means:
        T(x) = T_B(T_A(x))
             = T_B(x + A(x))
             = (x + A(x)) + B(x + A(x))
    
    The resulting displacement field C is:
        C(x) = T(x) - x = A(x) + B(x + A(x))
    
    Order of Application:
    --------------------
    **CRITICAL**: A is applied FIRST, then B is applied to the warped coordinates.
    
    Example: If you want to compose transforms [initial_warp, new_warp], call:
        compose_3d_fields_extrap(initial_warp, new_warp)
    
    This will apply initial_warp first, then new_warp, resulting in the full
    transformation chain.

    Implementation Details:
    ----------------------
    1. For each voxel x in the grid:
       - Compute warped coordinates: x_warped = x + A(x)
       - Interpolate B at x_warped to get B(x + A(x))
       - Compute composed displacement: C(x) = A(x) + B(x + A(x))
    
    2. Interpolation uses scipy.ndimage.map_coordinates with:
       - order=3: cubic B-spline interpolation (more accurate than linear)
       - mode='nearest': extrapolation at boundaries
       - prefilter=True: applies prefiltering for accurate B-spline interpolation
    
    3. The three vector components (x, y, z) are interpolated independently
       and in parallel using ThreadPoolExecutor.

    Parameters:
    ----------
    A : ndarray
        First displacement field, shape (X, Y, Z, 3), in voxel units.
        This transformation is applied FIRST.
    B : ndarray
        Second displacement field, shape (X, Y, Z, 3), in voxel units.
        This transformation is applied SECOND to the warped coordinates.
    interpolation_order : int, optional
        Order of B-spline interpolation (0=nearest, 1=linear, 3=cubic, 5=quintic).
        Default is 3 (cubic) for better accuracy.

    Returns:
    -------
    C : ndarray
        Composed displacement field, shape (X, Y, Z, 3), in voxel units.
        Represents the combined transformation: apply A first, then B.

    Notes:
    -----
    - Both A and B must be in voxel units (not mm). Use mm_to_vox_field() to
      convert NIfTI displacement fields from mm to voxel units before composition,
      then use vox_to_mm_field() to convert back.
    
    - The composition is NOT commutative: compose_3d_fields_extrap(A, B) ≠
      compose_3d_fields_extrap(B, A) in general.
    
    - For transform chains [T1, T2, T3] where you want to apply T1→T2→T3,
      you must compose pairwise:
      C12 = compose_3d_fields_extrap(T1, T2)  # T1 then T2
      C123 = compose_3d_fields_extrap(C12, T3)  # (T1∘T2) then T3
    
    - Higher interpolation orders (3 or 5) provide more accurate results than
      linear interpolation (order=1), especially for large deformations.
    """
    from scipy.ndimage import map_coordinates
    
    if A.shape != B.shape or A.shape[-1] != 3:
        raise ValueError("A and B must be (X, Y, Z, 3)")

    X, Y, Z, _ = A.shape
    i = np.arange(X, dtype=np.float64)  # Use float64 for better precision
    j = np.arange(Y, dtype=np.float64)
    k = np.arange(Z, dtype=np.float64)

    # Base grid and warped coords: x_warped = x + A(x)
    I, J, K = np.meshgrid(i, j, k, indexing='ij')
    
    # Compute warped coordinates with float64 precision
    Iw = I + A[..., 0].astype(np.float64)
    Jw = J + A[..., 1].astype(np.float64)
    Kw = K + A[..., 2].astype(np.float64)
    
    # Stack coordinates for map_coordinates (shape: (3, X*Y*Z))
    coords = np.stack([Iw.ravel(), Jw.ravel(), Kw.ravel()], axis=0)

    C = np.empty_like(A, dtype=np.float32)
    
    def interpolate_component(c):
        """Interpolate component c of B at the warped coordinates using cubic B-spline."""
        # Use map_coordinates with cubic B-spline interpolation
        Bw = map_coordinates(
            B[..., c].astype(np.float64),
            coords,
            order=interpolation_order,  # 3 = cubic B-spline
            mode='nearest',  # Extrapolation mode
            prefilter=True  # Apply B-spline prefiltering for accuracy
        ).reshape(X, Y, Z)
        
        # Compose: C_c(x) = A_c(x) + B_c(x + A(x))
        return c, (A[..., c] + Bw).astype(np.float32)
    
    # Use ThreadPoolExecutor for parallel interpolation of the 3 components
    with ThreadPoolExecutor(max_workers=3) as executor:
        results = executor.map(interpolate_component, range(3))
    
    for c, result in results:
        C[..., c] = result
    
    return C

def mm_to_vox_field(D_mm, affine):
    """
    Convert a displacement field in mm to voxel units.
    D_mm: (X,Y,Z,3) in mm; affine: (4,4) of the image grid.
    """
    # Use the linear part (rotation+scaling) to map vectors; ignore translation
    M = affine[:3, :3]
    Minv = np.linalg.inv(M)
    # Apply Minv to vectors (broadcast over grid)
    D_vox = np.tensordot(D_mm, Minv.T, axes=([3],[0]))  # -> (X,Y,Z,3)
    return D_vox

def vox_to_mm_field(D_vox, affine):
    M = affine[:3, :3]
    D_mm = np.tensordot(D_vox, M.T, axes=([3],[0]))
    return D_mm

def compose_warps(initial_warp, secondary_warp, out_path):
    # Load two vector fields (NIfTI) with last dim=3
    A_img = nib.load(initial_warp)
    B_img = nib.load(secondary_warp)
    A_mm = A_img.get_fdata(dtype=np.float32).squeeze()  # (X,Y,Z,3)
    B_mm = B_img.get_fdata(dtype=np.float32).squeeze()

    # Convert mm -> voxel units
    A_vox = mm_to_vox_field(A_mm, A_img.affine)
    B_vox = mm_to_vox_field(B_mm, B_img.affine)

    # Compose
    C_vox = compose_3d_fields_extrap(A_vox, B_vox)

    # Back to mm (optional)
    C_mm = vox_to_mm_field(C_vox, A_img.affine)

    # Ensure channel-last 3 components
    if C_mm.ndim == 4 and C_mm.shape[-1] == 3:
        pass
    elif C_mm.ndim == 4 and C_mm.shape[0] == 3:
        C_mm = np.moveaxis(C_mm, 0, -1)
    else:
        raise ValueError(f"Expected (X,Y,Z,3) or (3,X,Y,Z), got {C_mm.shape}")

    C_mm = np.ascontiguousarray(C_mm)

    X, Y, Z, _ = C_mm.shape
    # Make (X,Y,Z,1,3)
    data5 = C_mm[..., None, :]            # add a singleton 4th dim -> (X,Y,Z,1,3)

    hdr5 = nib.Nifti1Header()
    hdr5.set_data_dtype(np.float32)
    hdr5.set_xyzt_units('mm','sec')
    hdr5.set_intent('vector')

    hdr5['dim'][0] = 5                    # 5D image
    hdr5['dim'][1] = X
    hdr5['dim'][2] = Y
    hdr5['dim'][3] = Z
    hdr5['dim'][4] = 1
    hdr5['dim'][5] = 3                    # components = 3
    hdr5['pixdim'][4] = 1.0               # spacing for dim4
    hdr5['pixdim'][5] = 1.0               # spacing for dim5 (components), arbitrary

    img5 = nib.Nifti1Image(data5, affine=A_img.affine, header=hdr5)
    nib.save(img5, out_path)

def ants_linear_nonlinear_registration(
    fixed_file,
    moving_file,
    out_file=None,
    warp_file=None,
    affine_file=None,
    rev_warp_file=None,
    registration_method="SyNRA",
    initial_affine_file=None,
    initial_warp_file=None,
    interpolator="genericLabel",
    threads=DEFAULT_THREADS,
    fixed_image=None,
    initial_inverse_warp_file=None,
    disable_warp_composition=False,
    disable_inverse_warp_composition=False,
    **kwargs,
):
    """Perform linear (rigid + affine) and nonlinear registration using ANTsPy.

    This function performs registration between two images using ANTs' SyNRA transform,
    which includes both linear (rigid + affine) and nonlinear (SyN) components.

    Args:
        fixed_file (str): Path to the fixed/reference image.
        moving_file (str): Path to the moving image that will be registered.
        out_file (str, optional): Path where the registered image will be saved.
        warp_file (str, optional): Path to save the forward warp field.
        affine_file (str, optional): Path to save the forward affine transform.
        rev_warp_file (str, optional): Path to save the reverse warp field.
        rev_affine_file (str, optional): Path to save the reverse affine transform.
        registration_method (str): Registration method to use. Defaults to "SyNRA".
        initial_affine_file (str, optional): Path to initial affine transform.
        initial_warp_file (str, optional): Path to initial warp field.
        interpolator (str): Interpolation method. Defaults to "genericLabel".
        threads (int): Number of threads to use for registration. Defaults to all available cores.
        **kwargs: Additional arguments passed directly to ants.registration
                 Examples: verbose, grad_step, reg_iterations, etc.

    Returns:
        None: The function saves the registered image and transform files to disk.
    """
    if (
        not out_file
        and not warp_file
        and not affine_file
        and not rev_warp_file
    ):
        print(Fore.RED + "Error: No outputs specified." + Style.RESET_ALL)
        sys.exit(1)
    if not initial_warp_file and disable_warp_composition:
        print("Warning: --disable-warp-composition has no effect without --initial-warp-file.")
    if not initial_inverse_warp_file and disable_inverse_warp_composition:
        print("Warning: --disable-inverse-warp-composition has no effect without --initial-inverse-warp-file.")
    # Set ANTs/ITK thread count in environment variables
    os.environ["ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS"] = str(threads)
    os.environ["OMP_NUM_THREADS"] = str(threads)

    # Load images
    fixed = ants.image_read(fixed_file)
    moving = ants.image_read(moving_file)
    
    # Track temporary files for cleanup
    temp_files_to_cleanup = []
    
    # If initial transforms are provided, apply them first and register the warped image
    if initial_warp_file or initial_affine_file:
        print("Applying initial transforms to moving image...")
        
        # Build initial transform list (order matters: warp then affine)
        initial_transformlist = []
        if initial_warp_file:
            initial_transformlist.append(initial_warp_file)
        if initial_affine_file:
            initial_transformlist.append(initial_affine_file)
        
        # Apply initial transforms to moving image
        warped_moving = ants.apply_transforms(
            fixed=fixed,
            moving=moving,
            transformlist=initial_transformlist,
            interpolator=interpolator,
        )
        
        # Save temporarily warped moving image using tempfile
        temp_fd, temp_warped_file = tempfile.mkstemp(suffix='.nii.gz', prefix='temp_warped_moving_')
        os.close(temp_fd)  # Close the file descriptor as we'll write with ANTs
        ants.image_write(warped_moving, temp_warped_file)
        temp_files_to_cleanup.append(temp_warped_file)
        
        print(f"Initial transforms applied. Saved temporary warped image as {temp_warped_file}")
        print("Now performing new registration on pre-warped image...")
        
        # Register the warped moving image to fixed
        transforms = ants.registration(
            fixed=fixed,
            moving=warped_moving,
            type_of_transform=registration_method,
            initial_transform='Identity',  # No initial transform for this registration
            **kwargs,
        )
    else:
        # No initial transforms, just perform standard registration
        transforms = ants.registration(
            fixed=fixed,
            moving=moving,
            type_of_transform=registration_method,
            initial_transform=None,
            **kwargs,
        )
    
    
    # Save the registered moving image
    if out_file is not None:
        ants.image_write(transforms['warpedmovout'], out_file)
        print(f"Registration complete. Saved registered image as {out_file}")

    # Save forward warp
    if warp_file:
        if initial_warp_file:
            # Compose the new warp with the initial warp
            if disable_warp_composition:
                shutil.copyfile(transforms["fwdtransforms"][0], warp_file)
                print(f"Saved warp field as {warp_file} (composition disabled)")
            else:
                # Read the warp fields
                initial_warp_img = ants.image_read(initial_warp_file)
                secondary_warp_img = ants.image_read(transforms["fwdtransforms"][0])
                
                # Resample initial_warp to match secondary_warp grid
                # Use the first component of secondary_warp as target reference image
                target_ref = ants.split_channels(secondary_warp_img)[0]
                resampled_warp = resample_displacement_field_vectorially(
                    initial_warp_img, 
                    target_ref, 
                    interp='bSpline'  # Use linear for better stability
                )
                
                # Save resampled warp temporarily
                with tempfile.NamedTemporaryFile(suffix='.nii.gz', delete=False) as temp_resampled:
                    temp_resampled_path = temp_resampled.name
                    ants.image_write(resampled_warp, temp_resampled_path)
                    temp_files_to_cleanup.append(temp_resampled_path)
                
                # Compose: resampled_initial_warp (A) then secondary_warp (B)
                compose_warps(
                    temp_resampled_path,
                    transforms["fwdtransforms"][0],
                    warp_file
                )
                print(f"Composed and saved warp field as {warp_file}")
        else:
            shutil.copyfile(transforms["fwdtransforms"][0], warp_file)
            print(f"Saved warp field as {warp_file}")

    # Save forward affine
    if affine_file:
        if not initial_affine_file:
            shutil.copyfile(transforms["fwdtransforms"][1], affine_file)
            print(f"Saved affine transform as {affine_file}")
    
    # Save inverse warp
    if rev_warp_file:
        if initial_warp_file:
            if disable_inverse_warp_composition:
                shutil.copyfile(transforms["invtransforms"][1], rev_warp_file)
                print(f"Saved inverse warp field as {rev_warp_file} (composition disabled)")
            else:
                # Read the inverse warp fields
                initial_inv_warp_img = ants.image_read(initial_inverse_warp_file)
                secondary_inv_warp_img = ants.image_read(transforms["invtransforms"][1])
                
                # Resample initial_inverse_warp to match secondary_inverse_warp grid
                # Use the first component of secondary_inv_warp as target reference image
                target_ref = ants.split_channels(secondary_inv_warp_img)[0]
                resampled_inv_warp = resample_displacement_field_vectorially(
                    initial_inv_warp_img, 
                    target_ref, 
                    interp='bSpline'  # Use linear for better stability
                )
                
                # Save resampled inverse warp temporarily
                with tempfile.NamedTemporaryFile(suffix='.nii.gz', delete=False) as temp_resampled:
                    temp_resampled_path = temp_resampled.name
                    ants.image_write(resampled_inv_warp, temp_resampled_path)
                    temp_files_to_cleanup.append(temp_resampled_path)
                
                # Compose inverse warps: secondary_inverse_warp (A) then resampled_initial_inverse_warp (B)
                compose_warps(
                    transforms["invtransforms"][1],
                    temp_resampled_path,
                    rev_warp_file
                )
                print(f"Composed and saved inverse warp field as {rev_warp_file}")
        else:
            shutil.copyfile(transforms["invtransforms"][1], rev_warp_file)
            print(f"Saved inverse warp field as {rev_warp_file}")
    


    print("All specified outputs saved successfully.")
    
    # Clean up temporary files
    print("Cleaning up temporary files...")
    temp_files_to_delete = set(transforms['fwdtransforms'] + transforms['invtransforms'] + temp_files_to_cleanup)
    deleted_count = 0
    for temp_file in temp_files_to_delete:
        try:
            if os.path.exists(temp_file):
                os.remove(temp_file)
                deleted_count += 1
        except OSError as e:
            print(f"Warning: Could not remove temporary file {temp_file}: {e}")
    print(f"Successfully cleaned up {deleted_count} temporary files.")


def main():
    """Entry point for command-line use"""
    parser = argparse.ArgumentParser(description="Coregistration tool")

    # Required/standard arguments
    parser.add_argument("--fixed", required=True, help="Fixed/reference image file path")
    parser.add_argument("--moving", required=True, help="Moving image file path")
    parser.add_argument("--output", help="Output image file path")
    parser.add_argument(
        "--registration-method", default="SyNRA", help="Registration method"
    )
    parser.add_argument("--affine-file", help="Affine transformation file path")
    parser.add_argument("--warp-file", help="Warp field file path")
    parser.add_argument(
        "--inverse-warp-file", help="Inverse warp field file path"  # Standardized name
    )
    parser.add_argument(
        "--initial-affine-file", help="Initial affine transformation file path"
    )
    parser.add_argument("--initial-warp-file", help="Initial warp field file path")
    parser.add_argument("--disable-warp-composition", action="store_true", help="Disable warp composition with initial warp")
    parser.add_argument("--disable-inverse-warp-composition", action="store_true", help="Disable inverse warp composition with initial inverse warp")
    parser.add_argument(
        "--interpolator", help="Interpolator type", default="genericLabel"
    )
    
    # Add threads parameter with default value of all cores
    parser.add_argument(
        "--threads", 
        type=int, 
        default=DEFAULT_THREADS, 
        help=f"Number of threads to use (default: {DEFAULT_THREADS} - all cores)"
    )
    parser.add_argument(
        "--fixed-image", help="Original moving image, if provided, the warpfield will be resampled to this space."
    )
    parser.add_argument(
        "--initial-inverse-warp-file", help="Initial inverse warp field file path"
    )

    # Add common ANTs registration parameters
    parser.add_argument("--verbose", action="store_true", help="Enable verbose output")
    parser.add_argument(
        "--grad-step", type=float, default=0.2, help="Gradient step size (default: 0.2)"
    )
    parser.add_argument(
        "--flow-sigma",
        type=float,
        default=3,
    )
    parser.add_argument(
        "--syn-metric", default="mattes", help="Metric for SyN stage (default: mattes)"
    )
    parser.add_argument(
        "--syn-sampling",
        type=int,
        default=32,
        help="Sampling parameter for SyN metric (default: 32)",
    )
    parser.add_argument(
        "--total-sigma",
        type=float,
        default=0,
    )

    # More complex parameters that need special handling
    parser.add_argument(
        "--reg-iterations", help="SyN iterations, comma-separated (e.g., '40,20,0')"
    )
    parser.add_argument(
        "--aff-iterations",
        help="Affine iterations, comma-separated (e.g., '2100,1200,1200,10')",
    )
    parser.add_argument(
        "--aff-shrink-factors",
        help="Affine shrink factors, comma-separated (e.g., '6,4,2,1')",
    )
    parser.add_argument(
        "--aff-smoothing-sigmas",
        help="Affine smoothing sigmas, comma-separated (e.g., '3,2,1,0')",
    )
    parser.add_argument(
        "--aff-metric", default="mattes", help="Metric for affine stage (default: mattes)"
    )
    parser.add_argument("--aff-sampling", type=int, default=32, help="Sampling parameter for affine metric (default: 32)")
    parser.add_argument(
        "--random-seed", type=int, help="Random seed for reproducibility"
    )

    args = parser.parse_args()

    # Process tuple arguments from strings
    kwargs = {}

    if args.verbose:
        kwargs["verbose"] = True

    # Add standard numeric parameters
    for param in [
        "grad_step",
        "flow_sigma",
        "total_sigma",
        "aff_metric",
        "aff_sampling",
        "syn_metric",
        "syn_sampling",
    ]:
        param_value = getattr(args, param.replace("-", "_"))
        if param_value is not None:
            kwargs[param] = param_value

    # Convert comma-separated strings to tuples for complex parameters
    for param in [
        "reg_iterations",
        "aff_iterations",
        "aff_shrink_factors",
        "aff_smoothing_sigmas",
    ]:
        param_value = getattr(args, param.replace("-", "_"))
        if param_value:
            try:
                # Convert string "40,20,0" to tuple (40, 20, 0)
                kwargs[param] = tuple(int(x) for x in param_value.split(","))
            except ValueError:
                print(f"Error parsing {param}. Use comma-separated integers.")
                sys.exit(1)

    # Add random seed if specified
    if args.random_seed is not None:
        kwargs["random_seed"] = args.random_seed

    # Call the coregister function with all arguments
    ants_linear_nonlinear_registration(
        fixed_file=args.fixed,
        moving_file=args.moving,
        out_file=args.output,
        registration_method=args.registration_method,
        affine_file=args.affine_file,
        warp_file=args.warp_file,
        rev_warp_file=args.inverse_warp_file,  # Use standardized name 
        initial_affine_file=args.initial_affine_file,
        initial_warp_file=args.initial_warp_file,
        interpolator=args.interpolator,
        threads=args.threads,  # Pass threads parameter
        fixed_image=args.fixed_image,
        initial_inverse_warp_file=args.initial_inverse_warp_file,
        disable_warp_composition=args.disable_warp_composition,
        disable_inverse_warp_composition=args.disable_inverse_warp_composition,
        **kwargs,  # Pass all the extra parameters
    )


if __name__ == "__main__":
    main()
